from file_handlers.lumps.header import HeaderParser, HeaderPacker
from file_handlers.lumps.structure_parser import Struct
from file_handlers.lumps.l35 import Lump35Parser, Lump35Packer
from file_handlers.lumps.l40 import Lump40Parser, Lump40Packer
from pathlib import Path
from time import time
import lzma
import tempfile

class BSPLZMA():
    def __init__(self, fp):
        self.__header = HeaderParser(fp)
        self.__fp = fp

        self.__bsp_lzma = Struct("bsp_lzma_header",
                                 (("ID", "LZMA"),
                                  ("actualSize", "INTUL"),
                                  ("lzmaSize", "INTUL"),
                                  ("properties", "RAW/5")))

    def compress(self):
        header = self.__header.parse()

        if self.test_compressed():
            return
        
        ##we construct a new file so updating offsets is not quite so difficult.
        ##new offsets are calculated 'on the fly' as follows;
        ##we create a sorted list of tuples, containing the offset, size, and lump number of each lump
        ##now we operate on each offset sequentially.
        ##every offset that follows has the delta of size and new size subtracted
        ##the fourCC becomes the uncompressed size of the lump

        ##special cases:
        ##lump 40 is compressed as a pkzip archive would be (note: look at 7zip for this?)
        ## >> not necessary, we just compress each file after calculating crc32, change version to extract, compr. method
        ##gamelumps are individually compressed
        ##we must test to see if lumps are empty (offset, len = 0)

        offsets_list = []
        lnum = 0
        for lump in header["lump_t"]:
            offsets_list.append([lump["fileofs"], lump["filelen"], lnum])
            lnum += 1
        offsets_list = sorted(offsets_list) ##offsets are first so it sorts by them

        starttime = time()
        print("Repacking..", flush=True)
        with tempfile.TemporaryFile() as cfp:
            lump_t = [""]*64 ##initialise blank header array - to be updated later
            total_dsize = 0 ##the total size change to apply to offsets
            cfp.write(b"\x00"*1036)##reserving space at the start of the file for headers

            for item in offsets_list:                
                if item[1] == 0: ##unused lump case
                    lump_t[item[2]] = {"fileofs" : 0,"filelen" : 0,"version" : 0,"fourCC" : 0}
                    print(".", end="", flush=True)
                    
                elif item[2] != 40 and item[2] != 35:
                    size = item[1]
                    offset = item[0]
                    cfp.seek(0, 2)
                    offset = cfp.tell()
                    ## offset -= total_dsize
                    
                    self.__fp.seek(item[0])
                    lumpdata = lzma.compress(self.__fp.read(item[1]))
                    ########
                    ## look into using lzma stream compression, if memory issues are encountered ##
                    ########
                    lzmasize = len(lumpdata) ##lzma header size not added
                    
                    deltasize = size - lzmasize + 17 ##how much to adjust offsets by, incl. 17 bytes for lzma header
                    total_dsize += deltasize

                    lump_t[item[2]] = {"fileofs" : offset,
                                       "filelen" : lzmasize,
                                       "version" : 0,
                                       "fourCC" : size}
                    
                    ## we construct and insert the bsp lzma header here
                    new_lzma_header = {"ID" : "LZMA",
                                  "actualSize" : size,
                                  "lzmaSize" : lzmasize,
                                  "properties" : b"\x5d\x00\x00\x00\x01"}
                    ##properties may or may not always be \x5d\x00\x00\x00\x01;
                    ##I hope it is because I don't know how to calculate it if not
                    cfp.seek(0, 2)
                    self.__bsp_lzma.pack(cfp, new_lzma_header)
                    cfp.write(lumpdata)
                    cfp.seek(0, 2)

                        
                elif item[2] == 35: ##lump35 - gamelumps
                    ##gamelump offsets are adjusted
                    ##the length of each lump in the header remains the same
                    ##flags become 1 from 0
                    ##an extra dummy gamelump, with all fields = 0, is added
                    ##why and how this dummy gamelump is used is unknown to me, to figure out
                    ##--docs say it is used for offset calcs, but offset is zeroed
                    self.__fp.seek(header["lump_t"][35]["fileofs"])
                    l35parser = Lump35Parser(self.__fp)
                    l35packer = Lump35Packer(cfp)
                    l35header = l35parser.parse_header()
                    l35_dsize = 0 ##used for delta of size

                    cfp.seek(0, 2)
                    l35ofs = cfp.tell()
                    
                    for entry in l35header["dgamelump_t"]:
                        entry["flags"] = 1 ##seems to be 1 for repacked gamelumps
                        self.__fp.seek(entry["fileofs"])
                        if entry["fileofs"]:
                            lumpdata = lzma.compress(self.__fp.read(entry["filelen"]))
                            cfp.seek(0, 2) ##seek to end
                            glofs = cfp.tell() + 16 ##16 = length of dummy header added in later
                            new_header = {"ID" : "LZMA",
                                          "actualSize" : entry["filelen"],
                                          "lzmaSize" : len(lumpdata),##lzma header size not added
                                          "properties" : b"\x5d\x00\x00\x00\x01"}
                            self.__bsp_lzma.pack(cfp, new_header) 
                            cfp.write(lumpdata) ##we write the lzma header then the new data to the end of the file
                            deltasize = entry["filelen"] - len(lumpdata)
                            l35_dsize += deltasize
                                       
##                        entry["fileofs"] -= total_dsize
                        entry["fileofs"] = glofs ##gamelump offset
                        total_dsize += deltasize
                    ##dummy header
                    l35header["dgamelump_t"].append({"id" : b"\x00\x00\x00\x00",
                                                         "flags" : 0,
                                                         "version" : 0,
                                                         "fileofs" : 0,
                                                         "filelen" : 0})
                    l35header["lumpCount"] += 1

##                    "fileofs" : item[0] - total_dsize + l35_dsize
                    lump_t[item[2]] = {"fileofs" : l35ofs, ##add l35dsize otherwise totaldsize is incorrect
                                         "filelen" : item[1] - l35_dsize,
                                         "version" : 0,
                                         "fourCC" : 0} ##fourcc does NOT change for l35

    
                    cfp.seek(lump_t[item[2]]["fileofs"]) ##data is written already so we simply insert a header at the beginning of it
                    l35packer.pack_header(l35header, delta_ngamelumps = 1)
                    ##look into delta ngamelumps and its usage of overwrite bytes -> it seems to make a valid lump though so?
                    ##new l35 header with updated offsets.
                    ##for the purpose of delta_ngamelumps, see pack_header in l35.py for an explanation                    

                else: ##lump 40 - pakfile
                    l40parser = Lump40Parser(self.__fp)
                    l40packer = Lump40Packer(cfp)
                    l40data = l40parser.parse()
                    delta = 0
                    ## delta of size for this lump only
                    cfp.seek(0, 2)
                    l40ofs = cfp.tell()

                    ## only data that changes is stored data, compr. method, version to extract, compr. size
                    ## (and offsets)
                    ## so it is quite easy to compress l40 theoretically

                    nentries = l40data["zeocdr"]["nCentralDirectoryEntries_Total"]
                    for x in range(nentries):
                        ## we operate on every pair of entries in zfh, zlfh as they are linked
                        ## ---- zlfh ----
                        l40data["zlfh"][x]["versionNeededToExtract"] = 63
                        l40data["zlfh"][x]["compressionMethod"] = 14
                        
                        ## testing at first without lzma header on each file (doubt it's necessary though)
                        lfhd = l40data["zlfh"][x]["lfhData"]
                        lfhd = lzma.compress(lfhd) ## l40 parser set up already so no need to mess with offsets
                        csize = len(lfhd)
                        l40data["zlfh"][x]["lfhData"] = lfhd
                        l40data["zlfh"][x]["compressedSize"] = csize


                        ## ---- zfh ----
                        l40data["zfh"][x]["versionNeededToExtract"] = 63
                        l40data["zfh"][x]["compressionMethod"] = 14
                        l40data["zfh"][x]["relativeOffsetOfLocalHeader"] -= delta
                        l40data["zfh"][x]["compressedSize"] = csize


                        delta += (l40data["zlfh"][x]["uncompressedSize"] - csize)
                        ##we recalculate delta AFTER operations otherwise issues with the relative offset occur

                    l40data["zeocdr"]["startOfCentralDirOffset"] -= delta
                    
                    cfp.seek(0, 2)
                    lump_t[item[2]] = {"fileofs" : l40ofs,
                                         "filelen" : item[1] - delta,
                                         "version" : 0,
                                         "fourCC" : 0} ##fourcc does NOT change for l40

                    l40packer.pack_raw(l40data)
                ##end of file adjustments


            cfp.seek(0) ##create the bsp header
            new_header = {"ident" : "VBSP",
                          "version" : 20,
                          "lump_t" : lump_t,
                          "revision" : header["revision"]}
            header_packer = HeaderPacker(cfp)
##            for counter in range(len(new_header["lump_t"])):
##                if new_header["lump_t"][counter]["fileofs"] < 0 or new_header["lump_t"][counter]["filelen"] < 0:
##                    print(counter, new_header["lump_t"][counter])
            header_packer.pack(new_header)

            self.__fp.seek(0)
            self.__fp.truncate()
            cfp.seek(0)
            self.__fp.write(cfp.read())
            
            endtime = time()

            print("\n\nRepacking finished in %s seconds" %round(endtime-starttime, 2))
            
            

    def test_compressed(self):
        data = self.__header.parse()
        return not(data["lump_t"][0]["fourCC"] == 0)


